#include <ruby.h>
#include <taglib/tag_c.h>

typedef struct
{
    TagLib_File *file;
    TagLib_Tag *tag;
    const TagLib_AudioProperties *audio;
} tgFileData;

VALUE mTagLib;
VALUE eBadPath, eBadFile, eBadTag, eBadAudioProperties;
VALUE cFile;

static void
free_tgFileData(tgFileData *d)
{
    taglib_file_free(d->file);
    free(d);
}

TagLib_Tag *
file_tag(VALUE self)
{
    tgFileData *d;
    TagLib_Tag *tgTag;

    Data_Get_Struct(self, tgFileData, d);
    if (d->tag==NULL)
    {
        tgTag=taglib_file_tag(d->file);
        if (tgTag==NULL)
        	 rb_raise(eBadTag, "Bad tag");
        d->tag=tgTag;
	}
    return d->tag;
}

const TagLib_AudioProperties *
file_audio(VALUE self)
{
    tgFileData *d;
    const TagLib_AudioProperties *tgAudio;

    Data_Get_Struct(self, tgFileData, d);
    if (d->audio==NULL)
    {
        tgAudio=taglib_file_audioproperties(d->file);
        if (tgAudio==NULL)
            rb_raise(eBadAudioProperties, "Bad audio properties");
        d->audio=tgAudio;
    }
    return d->audio;
}

/*:nodoc:*/
VALUE
file_init(VALUE self, VALUE path)
{
    rb_iv_set(self, "@path", path);
    return self;
}

/*
call-seq: new(path)

Creates a new File object.

path: pathname of an audio file as a ruby string
*/
VALUE
file_new(VALUE self, VALUE path)
{
    TagLib_File *tgFile;
    tgFileData *data;
    VALUE rbData;
    VALUE argv[1];

    tgFile=taglib_file_new(StringValuePtr(path));
    if (tgFile==NULL)
        rb_raise(eBadPath, "Bad path");

    data=ALLOC(tgFileData);
    data->file=tgFile;
    data->tag=NULL;
    data->audio=NULL;
    rbData=Data_Wrap_Struct(cFile, 0, free_tgFileData, data);
    argv[0]=path;
    rb_obj_call_init(rbData, 1, argv);
    return rbData;
}

/*Save modifications to the audio file.*/ 
VALUE
file_save(VALUE self)
{
    tgFileData *data;

    Data_Get_Struct(self, tgFileData, data);
    if (taglib_file_save(data->file))
    {
        return Qtrue;
	}
    else
    {
        return Qfalse;
	}
}

/*Get track title*/
VALUE
file_get_title(VALUE self)
{
	VALUE str;

	str=rb_str_new2(taglib_tag_title(file_tag(self)));
	taglib_tag_free_strings;
    return str;
}

/*
call-seq: title=title

Set track title to title
 
title: a string
*/
VALUE
file_set_title(VALUE self, VALUE title)
{
    taglib_tag_set_title(file_tag(self), StringValuePtr(title));
    return title;
}

/*Get track artist*/
VALUE
file_get_artist(VALUE self)
{
	VALUE str;

	str=rb_str_new2(taglib_tag_artist(file_tag(self)));
	taglib_tag_free_strings;
    return str;
}

/*
call-seq: artist=artist

Set track artist to artist
 
artist: a string
*/
VALUE
file_set_artist(VALUE self, VALUE artist)
{
    taglib_tag_set_artist(file_tag(self), StringValuePtr(artist));
    return artist;
}

/*Get track album*/
VALUE
file_get_album(VALUE self)
{
	VALUE str;

	str=rb_str_new2(taglib_tag_album(file_tag(self)));
	taglib_tag_free_strings;
    return str;
}

/*
call-seq: album=album

Set track album to album
 
album: a string
*/
VALUE
file_set_album(VALUE self, VALUE album)
{
    taglib_tag_set_album(file_tag(self), StringValuePtr(album));
    return album;
}

/*Get track comment*/
VALUE
file_get_comment(VALUE self)
{
	VALUE str;

	str=rb_str_new2(taglib_tag_comment(file_tag(self)));
	taglib_tag_free_strings;
    return str;
}

/*
call-seq: comment=comment

Set track comment to comment
 
comment: a string
*/
VALUE
file_set_comment(VALUE self, VALUE comment)
{
    taglib_tag_set_comment(file_tag(self), StringValuePtr(comment));
    return comment;
}

/*Get track genre*/
VALUE
file_get_genre(VALUE self)
{
	VALUE str;

	str=rb_str_new2(taglib_tag_genre(file_tag(self)));
	taglib_tag_free_strings;
    return str;
}

/*
call-seq: genre=genre

Set track genre to genre
 
genre: a string
*/
VALUE
file_set_genre(VALUE self, VALUE genre)
{
    taglib_tag_set_genre(file_tag(self), StringValuePtr(genre));
    return genre;
}

/*Get track year*/
VALUE
file_get_year(VALUE self)
{
    return INT2FIX(taglib_tag_year(file_tag(self)));
}

/*
call-seq: year=year

Set track year to year
 
year: a Fixnum
*/
VALUE
file_set_year(VALUE self, VALUE year)
{
    taglib_tag_set_year(file_tag(self), NUM2INT(year));
    return year;
}

/*Get track number*/
VALUE
file_get_track(VALUE self)
{
    return INT2FIX(taglib_tag_track(file_tag(self)));
}

/*
call-seq: track=number

Set track number to number
 
number: a Fixnum
*/
VALUE
file_set_track(VALUE self, VALUE track)
{
    taglib_tag_set_track(file_tag(self), NUM2INT(track));
    return track;
}

/*Get track length in seconds*/
VALUE
file_get_length(VALUE self)
{
    return INT2NUM(taglib_audioproperties_length(file_audio(self)));
}

/*Get track bitrate in kb/s*/
VALUE
file_get_bitrate(VALUE self)
{
    return INT2NUM(taglib_audioproperties_bitrate(file_audio(self)));
}

/*Get track samplerate in Hz*/
VALUE
file_get_samplerate(VALUE self)
{
    return INT2NUM(taglib_audioproperties_samplerate(file_audio(self)));
}

/*Get the number of channels in the audio stream*/
VALUE
file_get_channels(VALUE self)
{
    return INT2NUM(taglib_audioproperties_channels(file_audio(self)));
}

void
Init_taglib2()
{
/* */
	mTagLib=rb_define_module("TagLib");

	eBadPath=rb_define_class_under(mTagLib, "BadPath", rb_eException);
    eBadFile=rb_define_class_under(mTagLib, "BadFile", rb_eException);
    eBadTag=rb_define_class_under(mTagLib, "BadTag", rb_eException);
    eBadAudioProperties=rb_define_class_under(mTagLib, "BadAudioProperties", rb_eException);

    cFile=rb_define_class_under(mTagLib, "File", rb_cObject);
    rb_define_singleton_method(cFile, "new", file_new, 1);
    rb_define_method(cFile, "initialize", file_init, 1);
    rb_define_method(cFile, "save", file_save, 0);
    rb_define_method(cFile, "title", file_get_title, 0);
    rb_define_method(cFile, "title=", file_set_title, 1);
    rb_define_method(cFile, "artist", file_get_artist, 0);
    rb_define_method(cFile, "artist=", file_set_artist, 1);
    rb_define_method(cFile, "album", file_get_album, 0);
    rb_define_method(cFile, "album=", file_set_album, 1);
    rb_define_method(cFile, "comment", file_get_comment, 0);
    rb_define_method(cFile, "comment=", file_set_comment, 1);
    rb_define_method(cFile, "genre", file_get_genre, 0);
    rb_define_method(cFile, "genre=", file_set_genre, 1);
    rb_define_method(cFile, "year", file_get_year, 0);
    rb_define_method(cFile, "year=", file_set_year, 1);
    rb_define_method(cFile, "track", file_get_track, 0);
    rb_define_method(cFile, "track=", file_set_track, 1);
    rb_define_method(cFile, "length", file_get_length, 0);
    rb_define_method(cFile, "bitrate", file_get_bitrate, 0);
    rb_define_method(cFile, "samplerate", file_get_samplerate, 0);
    rb_define_method(cFile, "channels", file_get_channels, 0);
}
